/*$Id: idxthread.c,v 1.24 1998/03/01 04:45:08 lindberg Exp $*/
/*$Name: ezmlm-idx-0324 $*/

/* idxthread.c contains routines to from the ezmlm-idx subject index build */
/* a structure of unique subjects as well as a table of messages with      */
/* pointers to the subject. This leads to information on message threads   */
/* arranged chronologically within the thread, and with the threads        */
/* arranged chronologically by the first message within the range.         */

#include <sys/types.h>
#include <sys/stat.h>
#include "error.h"
#include "alloc.h"
#include "str.h"
#include "stralloc.h"
#include "strerr.h"
#include "lock.h"
#include "idx.h"
#include "errtxt.h"
#include "substdio.h"
#include "fmt.h"
#include "readwrite.h"

static stralloc line = {0};		/* primary input */
static stralloc authline = {0};		/* second line of primary input */

static substdio ssindex;
static char indexbuf[1024];

static char strnum[FMT_ULONG];

struct stat st;

static void die_nomem(fatal)
char *fatal;
{
  strerr_die2x(111,fatal,ERR_NOMEM);
}

int fdlock;

/* NOTE: These do NOT prevent double locking */
static void lockup(fatal)
char *fatal;
{
  fdlock = open_append("lock");
  if (fdlock == -1)
    strerr_die2sys(111,fatal,ERR_OPEN_LOCK);
  if (lock_ex(fdlock) == -1) {
    close(fdlock);
    strerr_die2sys(111,fatal,ERR_OBTAIN_LOCK);
  }
}

static void unlock()
{
    close(fdlock);
}

static void newsub(psubt,subject,sublen,msg,fatal)
/* Initializes subentry pointed to by psubt, adds a '\0' to subject,    */
/* allocates space and copies in subject, and puts a pointer to it in   */
/* the entry. */
subentry *psubt;
char *subject;
unsigned int sublen;
unsigned long msg;
char *fatal;
{
  register char *cpfrom, *cpto;
  register unsigned int cpno;

  psubt->higher = (subentry *) 0;
  psubt->lower = (subentry *) 0;
  psubt->firstmsg = msg;
  if (!(psubt->sub = alloc ((sublen+1) * sizeof(char))))
    die_nomem(fatal);
  cpto = psubt->sub;
  cpno = sublen;
  cpfrom = subject;
  while (cpno--) *(cpto++) = *(cpfrom++);
  *cpto = '\0';
}

static void newauth(pmsgt,author,authlen,fatal)
/* Allocates space for author of length authlen+1 adding a terminal '\0' */
/* and puts the pointer in pmsgt->author.                                */

  msgentry *pmsgt;	/* entry for current message */
  char *author;         /* pointer to author string (not sz!) */
  unsigned int authlen;	/* lenth of author */
  char *fatal;		/* sz */

{
  register char *cpfrom, *cpto;
  register unsigned int cpno;

  if (!(pmsgt->author = alloc ((authlen+1) * sizeof(char))))
    die_nomem(fatal);
  cpto = pmsgt->author;
  cpno = authlen;
  cpfrom = author;
  while (cpno--) *(cpto++) = *(cpfrom++);
  *cpto = '\0';
}

void idx_mkthreads(pmsgtable,psubtable,msg_from,msg_to,msg_latest,locked,fatal)
/* Threads messages msg_from -> msg_to into pmsgtable & psubtable. When  */
/* reading the latest index file (containing msg_latest) it locks the    */
/* directory, unless it is already locked (as in digest creation).       */
/* msgtable has the subject number 1.. (0 if there is no subject match,  */
/* which should happen only if the subject index is corrupt.)            */

/* 19971107 Changed to deal with index files that are missing, or have   */
/* missing entries, not necessarily reflecting missing archive files.    */
/* This all to make ezmlm-get more robust to get maximal info out of     */
/* corrupted archives.                                                   */

  msgentry **pmsgtable;		/* table of message<->subject */
  subentry **psubtable;		/* subject no, len, str char * */
  unsigned long msg_from;	/* first message in range */
  unsigned long msg_to;		/* last message in range */
  unsigned long msg_latest;	/* latest message in archive (for locking) */
  int locked;			/* if already locked */
  char *fatal;			/* Program-specific */

{
  unsigned long idxlatest;	/* need to lock for this (last) index file */
  unsigned long msg;		/* current msg number */
  unsigned long endmsg;		/* max msg in this idx file */
  unsigned long tmpmsg;		/* index entry's msg number */
  unsigned long idx;		/* current index file no */
  unsigned long idxto;		/* index containing end of range */
  unsigned long ulmrange; 	/* total # of messages in range */
  char *subject;		/* subject on line */
  unsigned int sublen;		/* length of subject */
  int ffound;			/* msg subject was found in subtable */
  unsigned int pos,posa;
  unsigned long submax;		/* max subject num in subtable */
  subentry *psubnext;		/* points to next entry in subtable */
  subentry *psubt;		/* points to entry in subtable */
  int fd;			/* index file handle */
  int flagmissingindex;		/* current index file is missing */
  int flagauth;			/* read index entry has author info */
  int hasauth;			/* current msg's entry has author info */
  subentry dummy;
  msgentry *pmsgt;
  int res;
  int match;
  register msgentry *x, *y;

				/* a few unnecessary sanity checks */
  if (msg_to > msg_latest)
    msg_to = msg_latest;
  if ((ulmrange = msg_to - msg_from +1) <= 0)
    strerr_die2x(100,fatal,"Program error: bad range in idx_mkthreads");
  if (!(*pmsgtable = (msgentry *) alloc(ulmrange * sizeof(msgentry))))
        die_nomem(fatal);
  y = *pmsgtable;
  x = y + ulmrange;			/* clear */
  while (--x >= y) {
    x->subnum = 0;
    x->author = (char *) 0;
  }

  if (!(*psubtable = (subentry *) alloc((ulmrange+1) * sizeof(subentry))))
        die_nomem(fatal);

  idxlatest = msg_latest / 100;
  idxto = msg_to / 100;
  submax = 0;
  psubnext = *psubtable;	/* dummy node to get tree going. Basically, */
  psubt = &dummy;		/* assure that subject > psubt-sub and that */
  psubt->sub = "";		/* there is something to hold psubt->higher */
  psubt->higher = (subentry *) 0;
  psubt->lower = (subentry *) 0;

  for (idx = msg_from / 100; idx <= idxto; idx++) {
				/* make index file name */
    if (!stralloc_copys(&line,"archive/")) die_nomem(fatal);
    if (!stralloc_catb(&line,strnum,fmt_ulong(strnum,idx))) die_nomem(fatal);
    if (!stralloc_cats(&line,"/index")) die_nomem(fatal);
    if (!stralloc_0(&line)) die_nomem(fatal);
    if (!locked && idx == idxlatest)
      lockup(fatal);
    flagmissingindex = 0;
    fd = open_read(line.s);
    if (fd == -1) {
      if (errno == error_noent) {	/* this means the index is not here */
					/* but the lists is supposedly indexed*/
        flagmissingindex = 1;
      } else
        strerr_die4sys(111,fatal,ERR_OPEN,line.s,": ");
    } else
      substdio_fdbuf(&ssindex,read,fd,indexbuf,sizeof(indexbuf));

    msg = 100L * idx;			/* current msg# */
    if (!msg) msg = 1L;			/* for start to make msg > tmpmsg */
    tmpmsg = 0L;			/* msg number of read index line */
    endmsg = msg + 99L;			/* max msg in this index */
    if (endmsg > msg_to)		/* skip non-asked for subjects */
      endmsg = msg_to;
    for (; msg <= endmsg; msg++) {
      if (!flagmissingindex && (msg > tmpmsg)) {
        flagauth = 0;
        if (getln(&ssindex,&line,&match,'\n') == -1)
          strerr_die3sys(111,fatal,ERR_READ,"index: ");
        if (!match)
          flagmissingindex = 1;
        else {
          pos = scan_ulong(line.s,&tmpmsg);
          if (line.s[pos++] == ':') {
            if (getln(&ssindex,&authline,&match,'\n') == -1)
              strerr_die3sys(111,fatal,ERR_READ,"index: ");
            if (!match)
              flagmissingindex = 1;
            else
              flagauth = 1;
            pos++;
          }
        }
      }
      if (msg < msg_from)	/* Nothing before start of range */
        continue;
      if (msg == tmpmsg) {
        subject = line.s + pos;
        sublen = line.len - pos;
        hasauth = flagauth;
      } else {
        subject = TXT_NOINDEX;
        sublen = str_len(TXT_NOINDEX);
        hasauth = 0;
      } 
      ffound = 0;		/* search among already known subjects */
      while (!ffound) {
        res = str_diffn(psubt->sub,subject,20);
        if (res < 0) {
          if (psubt->higher)
            psubt = psubt->higher;
          else {
            newsub(psubnext,subject,sublen,msg,fatal);
            psubt->higher = psubnext;
            psubt = psubnext;
            psubnext++;
            ffound = 1;
          }
        } else if (res > 0) {
          if (psubt->lower)
            psubt = psubt->lower;
          else {
            newsub(psubnext,subject,sublen,msg,fatal);
            psubt->lower = psubnext;
            psubt = psubnext;
            psubnext++;
            ffound = 1;
          }
        } else {
          ffound = 1;
        }
      }
 	           		/* first subnum =1 (=0 is empty for thread) */
      pmsgt = (*pmsgtable) + msg - msg_from;
      pmsgt->subnum = (unsigned int) (psubt - *psubtable + 1);
      pmsgt->author = (char *) 0;
      if (hasauth) {
        posa = byte_chr(authline.s,authline.len,';');
        if (authline.len > posa + 3)
          newauth(pmsgt,authline.s+posa+2,authline.len-posa-3,fatal);
      }


      psubt = *psubtable;	/* setup psubt. Done here rather than before */
				/* when(!ffound), so that we can start off    */
				/* the dummy node. */
    }
    if (fd != -1)
      close(fd);
    if (!locked && idx == idxlatest)
      unlock();			/* 'locked' refers to locked before calling */
  }
  psubnext->sub = (char *) 0;		/* end of table marker */
}


void idx_mkthread(pmsgtable,psubtable,msg_from,msg_to,msg_master,
		msg_latest,locked,fatal)
/* Works like idx_mkthreads, except that it finds the subject for message   */
/* msg_master, then identifies messages in the range that have the same     */
/* subject. msgtable entries with subject 0 do not match, with '1' do match.*/

msgentry **pmsgtable;		/* pointer to table of message<->subject */
subentry **psubtable;		/* ptr to tbl of subject no, len, str char * */
unsigned long msg_from;		/* first message in range */
unsigned long msg_to;		/* last message in range */
unsigned long msg_latest;	/* latest message in archive (for locking) */
unsigned long msg_master;	/* master message for single thread, else 0*/
int locked;			/* if already locked */
char *fatal;			/* Program-specific */

{
  unsigned long idxlatest;	/* need to lock for this (last) index file */
  unsigned long idxto;		/* index for last msg in range */
  unsigned long idx;		/* current index file no */
  unsigned long msg;		/* index entry's msg number */
  unsigned long ulmrange; 	/* total # of messages in range */
  subentry *psubt;		/* points to last entry in subtable */
  int ffound;			/* msg subject was found in subtable */
  int flagauth;			/* there is author info */
  int firstfound = 1;		/* = 1 until first message in thread found */
  unsigned int pos;
  int fd;			/* index file handle */
  int match;
  msgentry *pmsgt;
  register msgentry *x,*y;

  if ((ulmrange = msg_to - msg_from +1) <= 0)
    strerr_die2x(100,fatal,"Program error: bad range in idx_mkthreads");
  if (!(*pmsgtable = (msgentry *) alloc(ulmrange * sizeof(msgentry))))
         die_nomem(fatal);
  y = *pmsgtable;
  x = y + ulmrange;
  while (--x >= y) {
    x->subnum = 0;
    x->author = (char *) 0;
  }
    
  if (!(*psubtable = (subentry *) alloc(2 * sizeof(subentry))))
          die_nomem(fatal);

  idxlatest = msg_latest / 100;
  idxto = msg_to / 100;
  idx = msg_master / 100;	/* index for master subject */

	/* Get master subject */
  if (!stralloc_copys(&line,"archive/")) die_nomem(fatal);
  if (!stralloc_catb(&line,strnum,fmt_ulong(strnum,idx))) die_nomem(fatal);
  if (!stralloc_cats(&line,"/index")) die_nomem(fatal);
  if (!stralloc_0(&line)) die_nomem(fatal);
  ffound = 0;
  if (!locked && idx == idxlatest)
    lockup(fatal);
  fd = open_read(line.s);
  psubt = *psubtable;
  if (fd == -1) {
    if (errno != error_noent)
      strerr_die4sys(111,fatal,ERR_OPEN,line.s,": ");
    else
      strerr_die2x(100,fatal,ERR_NOINDEX);
  } else {
    substdio_fdbuf(&ssindex,read,fd,indexbuf,sizeof(indexbuf));
    for(;;) {
      if (getln(&ssindex,&line,&match,'\n') == -1)
          strerr_die3sys(111,fatal,ERR_OPEN,"index: ");
      if (!match)
        break;
      pos=scan_ulong(line.s,&msg);
      if (line.s[pos++] == ':') {	 /* marker for author info */
        pos++;
        flagauth = 1;
      } else
        flagauth = 0;
      if (msg == msg_master) {
        newsub(psubt,line.s+pos,line.len-pos,msg,fatal);
					/* need to update msg later! */
        ffound = 1;
        break;
      }
      if (flagauth) {			/* skip author line */
        if (getln(&ssindex,&line,&match,'\n') == -1)
          strerr_die3sys(111,fatal,ERR_OPEN,"index: ");
      if (!match)
        break;
      }
    }
    close(fd);
  }
  if (!locked && idx == idxlatest)
    unlock();
  if (!ffound)
      strerr_die2x(100,fatal,ERR_NOINDEX);
  for (idx = msg_from / 100; idx <= idxto; idx++) {
		/* make index file name */
    if (!stralloc_copys(&line,"archive/")) die_nomem(fatal);
    if (!stralloc_catb(&line,strnum,fmt_ulong(strnum,idx))) die_nomem(fatal);
    if (!stralloc_cats(&line,"/index")) die_nomem(fatal);
    if (!stralloc_0(&line)) die_nomem(fatal);
    if (!locked && idx == idxlatest)
      lockup(fatal);
    fd = open_read(line.s);
    if (fd == -1) {
      if (errno != error_noent)
        strerr_die4sys(111,fatal,ERR_OPEN,line.s,": ");
    } else {
      substdio_fdbuf(&ssindex,read,fd,indexbuf,sizeof(indexbuf));
      for(;;) {
        if (getln(&ssindex,&line,&match,'\n') == -1)
          strerr_die3sys(111,fatal,ERR_READ,"index: ");
        if (!match)
          break;
        pos=scan_ulong(line.s,&msg);
        if (line.s[pos++] == ':') {
          pos++;
          flagauth = 1;
          if (getln(&ssindex,&authline,&match,'\n') == -1)
            strerr_die3sys(111,fatal,ERR_READ,"index: ");
          if (!match)
            break;
        } else
          flagauth = 0;
        if (msg < msg_from)	/* Nothing before start of range */
          continue;
        if (msg > msg_to)	/* Don't do anything after range */
          break;
        if (!str_diffn(psubt->sub,line.s+pos,20)) {
          pmsgt = *pmsgtable + msg - msg_from;
          if (firstfound) {	/* update to first message with this subj */
            psubt->firstmsg = msg;
            firstfound = 0;
          }
          pmsgt->subnum = 1;
          if (flagauth) {
            pos = byte_chr(authline.s,authline.len,';');
            if (authline.len > pos + 3)
              newauth(pmsgt,authline.s+pos+2,authline.len-pos-3,fatal);
          }
        }
      }
      close(fd);
    }
    if (!locked && idx == idxlatest)
      unlock();
  }
  ++psubt;
  psubt->sub = (char *) 0;	/* end of table marker */
}

void idx_mklist(pmsgtable,psubtable,msg_from,msg_to,fatal)
/* Like mkthreads, except that it works without a subject index. The result */
/* is just a dummy subject and a sequential list of messages. This to allow */
/* use of the same routines when creating digest from lists that have no    */
/* subject index (for whatever reason). */
msgentry **pmsgtable;		/* pointer to table of message<->subject */
subentry **psubtable;		/* ptr to tbl of subject no, len, str char * */
unsigned long msg_from;		/* first message in range */
unsigned long msg_to;		/* last message in range */
char *fatal;			/* Program-specific */
{
  unsigned long ulmrange;
  register msgentry *x,*y;
  subentry *psubt;

  if ((ulmrange = msg_to - msg_from +1) <= 0)
    strerr_die2x(111,fatal,"bad range in idx_mkthreads :");

  if (!(*pmsgtable = (msgentry *) alloc(ulmrange * sizeof(msgentry))))
         die_nomem(fatal);

  y = *pmsgtable;
  x = y + ulmrange;
  while (--x >= y) {
    x->subnum = 1;
    x->author = (char *) 0;
  }
    
  if (!(*psubtable = (subentry *) alloc(2 * sizeof(subentry))))
          die_nomem(fatal);
  psubt = *psubtable;
  newsub(psubt,TXT_NOINDEX,str_len(TXT_NOINDEX),msg_from,fatal);
  ++psubt;
  psubt->sub = (char *) 0;
}
  
void idx_destroythread(msgtable,nomsg,subtable)
/* Frees space allocated by idxthread routines. */
msgentry *msgtable; subentry *subtable;
unsigned long nomsg;		/* number of msgtable entries */
{
  subentry *psubt;
  msgentry *pmsgt;
 
  psubt = subtable;		/* free subjects */
  while(psubt->sub) {
    alloc_free(psubt->sub);
    psubt++;
  }

  pmsgt = msgtable + nomsg;	/* free authors */
  while(nomsg--)
    alloc_free(pmsgt->author);	/* NULL is ok -> NOP */

  alloc_free(subtable);		/* free subtable */
  alloc_free(msgtable);		/* free msgtable */
  subtable = (subentry *) 0;
  msgtable = (msgentry *) 0;
}
